import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import javax.jms.*;

class SharedConnection implements Connection {

    private final Connection connection;
    private final SessionPools pools;
    private final List<SharedSession> sessions = new LinkedList<SharedSession>();
    private final Hashtable<String,TemporaryQueue> tmpQueues =
              new Hashtable<String,TemporaryQueue>();
    private final Hashtable<String,TemporaryTopic> tmpTopics =
              new Hashtable<String,TemporaryTopic>();
    private final Object clientIDLock = new Object();
    private String clientID;
    private boolean closed;

    SharedConnection(final Connection connection, final SessionPools pools, final String clientID) {
        this.connection = connection;
        this.pools = pools;
        this.clientID = clientID;
    }

    public Session createSession(final boolean transacted, final int acknowledgeMode)
            throws JMSException {
        synchronized(sessions) {
            if (closed) {
                throw new javax.jms.IllegalStateException("Connection has been closed");
            }
            final Pool<SessionHolder> pool;
            if (transacted) {
                pool = pools.SESSION_TRANSACTED;
            } else {
                switch (acknowledgeMode) {
                    case Session.AUTO_ACKNOWLEDGE: pool = pools.AUTO_ACKNOWLEDGE; break;
                    case Session.CLIENT_ACKNOWLEDGE: pool = pools.CLIENT_ACKNOWLEDGE; break;
                    case Session.DUPS_OK_ACKNOWLEDGE: pool = pools.DUPS_OK_ACKNOWLEDGE; break;
                    default: throw new JMSException("Unknown acknowledge mode: "+acknowledgeMode);
                }
            }
            SessionHolder sessionHolder = pool.get();
            if (sessionHolder == null) {
                sessionHolder = new SessionHolder(connection.createSession(transacted, acknowledgeMode));
            }
            final SharedSession session = new SharedSession(sessionHolder, pool, tmpQueues, tmpTopics);
            sessions.add(session);
            return session;
        }
    }

    public String getClientID() throws JMSException {
        return connection.getClientID();
    }

    public void setClientID(final String clientID) throws JMSException {
        synchronized(clientIDLock) {
            if (this.clientID == null) {
                connection.setClientID(clientID);
                this.clientID = clientID;
            } else {
                if (!this.clientID.equals(clientID)) {
                    throw new JMSException("Cannot change ClientID from '"+
                            this.clientID+"' to '"+clientID+"' on a shared connection!");
                }
            }
        }
    }

    public ConnectionMetaData getMetaData() throws JMSException {
        return connection.getMetaData();
    }

    public ExceptionListener getExceptionListener() throws JMSException {
        return connection.getExceptionListener();
    }

    public void setExceptionListener(final ExceptionListener listener) throws JMSException {
        // clients should call getExceptionListener() first and then
        // call the original listener from their own listener
        connection.setExceptionListener(listener);
    }

    public void start() throws JMSException {
        // always started
    }

    public void stop() throws JMSException {
        throw new JMSException("SharedConnection should not be stopped!");
    }

    public void close() throws JMSException {
        try {
            synchronized(sessions) {
                if (closed) return;
                closed = true;
                if (!sessions.isEmpty()) {
                    try {
                        closeNextSession(sessions.iterator());
                    } finally {
                        sessions.clear();
                    }
                }
            }
        } finally {
            try {
                synchronized(tmpQueues) {
                    if (!tmpQueues.isEmpty()) {
                        try {
                            deleteNextTemporaryQueue(tmpQueues.elements());
                        } finally {
                            tmpQueues.clear();
                        }
                    }
                }
            } finally {
                synchronized(tmpTopics) {
                    if (!tmpTopics.isEmpty()) {
                        try {
                            deleteNextTemporaryTopic(tmpTopics.elements());
                        } finally {
                            tmpTopics.clear();
                        }
                    }
                }
            }
        }
    }

    private static void closeNextSession(final Iterator<SharedSession> it) throws JMSException {
        if (it.hasNext()) {
            try {
                it.next().close();
            } finally {
                closeNextSession(it);
            }
        }
    }

    private static void deleteNextTemporaryQueue(final Enumeration<TemporaryQueue> it)
            throws JMSException {
        if (it.hasMoreElements()) {
            try {
                it.nextElement().delete();
            } finally {
                deleteNextTemporaryQueue(it);
            }
        }
    }

    private static void deleteNextTemporaryTopic(final Enumeration<TemporaryTopic> it)
            throws JMSException {
        if (it.hasMoreElements()) {
            try {
                it.nextElement().delete();
            } finally {
                deleteNextTemporaryTopic(it);
            }
        }
    }

    public ConnectionConsumer createConnectionConsumer(final Destination destination,
            final String messageSelector, final ServerSessionPool sessionPool,
            final int maxMessages) throws JMSException {
        throw new UnsupportedOperationException();
    }

    public ConnectionConsumer createDurableConnectionConsumer(final Topic destination,
            final String subscriptionName, final String messageSelector,
            final ServerSessionPool sessionPool, final int maxMessages) throws JMSException {
        throw new UnsupportedOperationException();
    }

}
