import java.util.HashMap;
import java.util.Iterator;
import java.util.concurrent.atomic.AtomicBoolean;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.JMSException;

public class SharedConnectionFactory implements ConnectionFactory {

    private final ConnectionFactory connectionFactory;
    private final String clientID;
    private final HashMap<String,ConnectionHolder> connections =
              new HashMap<String,ConnectionHolder>();

    public SharedConnectionFactory(final ConnectionFactory connectionFactory) {
        this(connectionFactory, null);
    }

    public SharedConnectionFactory(final ConnectionFactory connectionFactory,
            final String clientID) {
        if (connectionFactory == null) {
            throw new NullPointerException("ConnectionFactory cannot be null!");
        }
        this.connectionFactory = connectionFactory;
        this.clientID = clientID;
    }

    private static class ConnectionHolder {
        public final Connection connection;
        public final SessionPools pools = new SessionPools();
        public ConnectionHolder(final Connection connection) {
            this.connection = connection;
        }
    }

    private static void closeNextConnection(final Iterator<ConnectionHolder> it)
            throws JMSException {
        if (it.hasNext()) {
            try {
                it.next().connection.close();
            } finally {
                closeNextConnection(it);
            }
        }
    }

    public void close() throws JMSException {
        synchronized(connections) {
            if (!connections.isEmpty()) {
                try {
                    closeNextConnection(connections.values().iterator());
                } finally {
                    connections.clear();
                }
            }
        }
    }

    public Connection createConnection() throws JMSException {
        return createConnection(null, null);
    }

    public Connection createConnection(final String username, final String password)
            throws JMSException {
        ConnectionHolder holder;
        synchronized(connections) {
            holder = connections.get(username);
            if (holder == null) {
                holder = new ConnectionHolder(prepareConnection(username, password));
                connections.put(username, holder);
            }
        }
        return new SharedConnection(holder.connection, holder.pools, clientID);
    }

    private Connection prepareConnection(final String username, final String password)
            throws JMSException {
        final Connection connection = username == null ? connectionFactory.createConnection() :
                                      connectionFactory.createConnection(username, password);
        if (connection == null) {
            String message = "ConnectionFactory created null Connection";
            if (username != null) message+=" for username '"+username+"'";
            throw new JMSException(message);
        }
        boolean done = false;
        try {
            if (clientID != null) connection.setClientID(clientID);
            connection.start();
            connection.setExceptionListener(new javax.jms.ExceptionListener() {
                private final AtomicBoolean closing = new AtomicBoolean();
                public void onException(JMSException jmsException) {
                    if (closing.getAndSet(true)) return;
                    try {
                        connection.close();
                    } catch (JMSException ignored) {
                        // one call to close() should be enough
                    } finally {
                        closing.set(false);
                    }
                    synchronized(connections) {
                        final ConnectionHolder holder = connections.get(username);
                        if (holder != null && holder.connection == connection) {
                            connections.remove(username);
                        }
                    }
                }
            });
            done = true;
        } finally {
            if (!done) {
                // an exception occured
                try {
                    connection.close();
                } catch (Throwable ignored) {
                    // throw the previous exception instead
                }
            }
        }
        return connection;
    }

}
