import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageListener;

class SharedMessageConsumer implements MessageConsumer, MessageListener {

    private final MessageConsumer consumer;
    private final Object onMessageLock = new Object();
    private MessageListener listener;
    private long listenerSleep = SLEEP_INIT;
    private boolean closed;
    private static final long SLEEP_INIT = 1;
    private static final long SLEEP_MAX = 1024;

    SharedMessageConsumer(final MessageConsumer consumer) {
        if (consumer == null) {
            throw new NullPointerException("MessageConsumer cannot be null!");
        }
        this.consumer = consumer;
    }

    private void checkClosed() throws javax.jms.IllegalStateException {
        if (closed) throw new javax.jms.IllegalStateException("MessageConsumer has been closed");
    }

    public void onMessage(final Message msg) {
        synchronized(onMessageLock) {
            if (listener == null) {
                /* This means that setMessageListener(null) is going to be called soon,
                 * but it has not been called yet. So lets do a bit of magic to prevent
                 * RuntimeException from being thrown many times on the same Message -
                 * that might cause the JMS provider to defer delivery of that Message,
                 * which would not be nice.
                 */
                try {
                    Thread.sleep(listenerSleep);
                } catch (InterruptedException e) {
                    try {
                        Thread.sleep(listenerSleep);
                        Thread.currentThread().interrupt();
                    } catch (InterruptedException e2) {
                        Thread.currentThread().interrupt();
                    }
                }
                listenerSleep = listenerSleep << 1;
                if (listenerSleep > SLEEP_MAX) {
                    listenerSleep = SLEEP_MAX;
                }
                throw new RuntimeException(
            "MessageConsumer closed, preserving Message by throwing RuntimeException from MessageListener");
            }
            listener.onMessage(msg);
        }
    }

    public synchronized void setMessageListener(final MessageListener newListener) throws JMSException {
        checkClosed();
        boolean deleteMessageListener = false;
        synchronized(onMessageLock) {
            if (newListener == null) {
                if (listener != null) {
                    listenerSleep = SLEEP_INIT;
                    listener = null;
                    deleteMessageListener = true;
                    /* Calling consumer.setMessageListener(null) here (while holding
                     * onMessageLock) could cause a deadlock if the consumer would
                     * like to wait for its current MessageListener#onMessage(Message)
                     * to complete.
                     */
                }
            } else {
                if (listener == null) {
                    consumer.setMessageListener(this);
                }
                listener = newListener;
            }
        }
        if (deleteMessageListener) {
            consumer.setMessageListener(null);
        }
    }

    public synchronized MessageListener getMessageListener() throws JMSException {
        checkClosed();
        return consumer.getMessageListener();
    }

    public synchronized String getMessageSelector() throws JMSException {
        checkClosed();
        return consumer.getMessageSelector();
    }

    public synchronized Message receive() throws JMSException {
        // this should not be used because it could block closing
        checkClosed();
        return consumer.receive();
    }

    public synchronized Message receive(long timeout) throws JMSException {
        checkClosed();
        return consumer.receive(timeout);
    }

    public synchronized Message receiveNoWait() throws JMSException {
        checkClosed();
        return consumer.receiveNoWait();
    }

    public synchronized void close() throws JMSException {
        if (closed) return;
        setMessageListener(null);
        closed = true;
    }

}
